(defun neighbours (pt lim)
  (destructuring-bind (x y xs ys) `(,@pt ,@lim)
		      (loop for dx from (- x 1) to (+ x 1) appending
			    (loop for dy from (- y 1) to (+ y 1)
				  if (and (>= dx 0) (< dx xs) (>= dy 0) (< dy ys))
				  collect (list dx dy)))))


(defun read-board (fname)
  (with-open-file (the-board fname)
		  (let ((blist (loop for l = (read-line the-board nil) 
				     while l collect l)))
		    (make-array (list (length (car blist)) (length blist))
				:initial-contents (mapcar (lambda (m) (concatenate 'list m)) blist)))))

(setq r '((#\0 0) (#\1 1) (#\2 2) (#\3 3) (#\4 4) (#\5 5) (#\6 6) (#\7 7) (#\8 8) (#\9 9)))

(defun make-rules (board)
  (let ((pd (array-dimensions board)))
    (destructuring-bind (xs ys) pd
			(loop for x from 0 to (- xs 1)
			      appending (loop for y from 0 to (- ys 1)
					      unless (equal (aref board x y) #\.)
					      collect (list (neighbours (list x y) pd)
							    (cadr (assoc (aref board x y) r))))))))

(defmacro filter (function list)
  (let ((a (gensym)))
    `(loop for ,a in ,list if (,function ,a) collect ,a)))

(defmacro aref* (board loc) `(aref ,board (car ,loc) (cadr ,loc)))


(defun solve-board (boardname)
  (setq pz (read-board boardname))
  (setq pr (make-rules pz))
  (setq xboard (make-array (array-dimensions pz) :initial-element -1))
  
  (setq changed t)
  (do () ((eq changed nil) xboard)
      (setq changed nil)
      (setq oxboard xboard)
      (dolist (rule pr)
	(cond
; honest zero => all are zero
	 ((= (cadr rule) 0) 
	  (progn (setq changed t)
		 (mapcar (lambda (u) (setf (aref* xboard u) 0)) (car rule))))
; fully satisfied
	 ((= (cadr rule) (length (car rule)))
	  (progn (setq changed t)
		 (mapcar (lambda (u) (setf (aref* xboard u) 1)) (car rule))))))
      
; change the rule list to account for the entries on the board
      (setq pr
	    (let ((rxacc nil))
	      (dolist (rule pr rxacc)
		(let ((ix (cadr rule))
		      (dl (mapcar (lambda (u) (aref* xboard u)) (car rule)))
		      (unassigned (filter (lambda (u) (= (aref* xboard u) -1)) (car rule))))
		  (let ((zeroes (count 0 dl)) 
			(ones (count 1 dl)))
		    (if unassigned
			(setq rxacc (cons (list unassigned (- ix ones)) rxacc))))))))))

(print (solve-board "puzzle1.txt"))